var id = 0
  , util = require('util')
  , Constants = require('jsr-constants')
  , CLIENT_TYPE = Constants.CLIENT_TYPE
  , CLIENT_FLAGS = Constants.CLIENT_FLAGS
  , Socket = require('net').Socket
  , keys = [
    'id',
    'addr',
    'fd',
    'age',
    'idle',
    'flags',
    'db',
    'sub',
    'psub',
    'multi',
    'cmd',
    'qbuf',
  ]

/**
 *  Encapsulates client connection state.
 */
function Client(socket) {

  // flag whether an external or internal client socket
  this._tcp = (socket instanceof Socket);

  // index into the connections array
  this._index = -1;

  // file descriptor
  this._fd = socket._handle.fd;

  // remote ip address
  this._remoteAddress = socket.remoteAddress;

  // remote port
  this._remotePort = socket.remotePort;

  // full remote address
  this._addr =
    (this._remoteAddress && this._remotePort)
      ? this._remoteAddress + ':' + this._remotePort : null;

  // connection id (2^53)
  this._id = ++id;

  // connection name used by setname/getname
  this._name = null;

  // flag used to mark clients as authenticated
  this._authenticated = false;

  // current db for this connection
  this.db = 0;

  // transaction reference for multi blocks
  this._transaction = null;

  // map of keys being watched by this connection
  this._watched = {};

  // monitor slave mode
  this._monitor = false;

  // the timestamp (ms) when this connection was established
  this._connected = Date.now();

  // last command played
  this._cmd = null;

  // buffer for queries
  this._qbuf = new Buffer(0);

  // time when the last command was received, used
  // to get idle time
  this._last = Date.now();

  // encapsulates pubsub state for this client
  this._pubsub = null;
}

/**
 *  Get an object of information about this client.
 */
function getObject() {
  var o = {}
    , i
    , fields = arguments.length
      ? Array.prototype.slice.call(arguments, 0) : keys;
  for(i = 0;i < fields.length;i++) {
    o[fields[i]] = this[fields[i]];
  }
  return o;
}

/**
 *  Get a string representing this client.
 */
function toString() {
  var s = ''
    , k
    , o = this.getObject.apply(this, arguments);
  for(k in o) {
    if(s) s += ' ';
    s += util.format('%s=%s', k, o[k]);
  }
  return s + '\n';
}

/**
 *  Determine if this client is a slave.
 */
function isSlave() {
  return this.type === CLIENT_TYPE.SLAVE;
}

// fd
Object.defineProperty(Client.prototype, 'fd', {
  get: function getFd() {
    return this._fd;
  }
})

// flags
Object.defineProperty(Client.prototype, 'flags', {
  get: function getFlags() {
    var flags = CLIENT_FLAGS.NONE;
    if(this.monitor) {
      flags = CLIENT_FLAGS.MONITOR;
    }else if(this._transaction) {
      flags = CLIENT_FLAGS.MULTI_EXEC;
      if(this._transaction.conflict) {
        flags += CLIENT_FLAGS.CONFLICT;
      }
    }
    return flags;
  }
})

// id
Object.defineProperty(Client.prototype, 'id', {
  get: function getId() {
    return this._id;
  }
})

// client index
Object.defineProperty(Client.prototype, 'index', {
  get: function getIndex() {
    return this._index;
  },
  set: function setIndex(index) {
    this._index = index;
  }
})

// addr
Object.defineProperty(Client.prototype, 'addr', {
  get: function getAddr() {
    return this._addr;
  }
})

// cmd
Object.defineProperty(Client.prototype, 'cmd', {
  get: function getCommand() {
    return this._cmd;
  },
  set: function setCommand(cmd) {
    this._cmd = cmd;
    this._last = Date.now();
  }
})

// qbuf
Object.defineProperty(Client.prototype, 'qbuf', {
  get: function getCommand() {
    return this._qbuf.length;
  }
})

// multi - number of queued transactions
Object.defineProperty(Client.prototype, 'multi', {
  get: function getMulti() {
    return this._transaction ? this._transaction.queue.length : -1;
  }
})

// age in seconds
Object.defineProperty(Client.prototype, 'age', {
  get: function getAge() {
    return Math.round((Date.now() - this._connected) / 1000);
  }
})

// idle time in seconds
Object.defineProperty(Client.prototype, 'idle', {
  get: function getIdle() {
    return Math.round((Date.now() - this._last) / 1000);
  }
})

// name
Object.defineProperty(Client.prototype, 'name', {
  get: function getName() {
    return this._name;
  },
  set: function setName(name) {
    this._name = name;
  }
})

// authenticated
Object.defineProperty(Client.prototype, 'authenticated', {
  get: function getAuthenticated() {
    return this._authenticated;
  },
  set: function setAuthenticated(auth) {
    this._authenticated = auth;
  }
})

// db
Object.defineProperty(Client.prototype, 'db', {
  get: function getDatabase() {
    return this._db;
  },
  set: function setDatabase(db) {
    this._db = db;
  }
})

// sub
Object.defineProperty(Client.prototype, 'sub', {
  get: function getSub() {
    if(!this._pubsub || !this._pubsub.channels) return 0;
    return this._pubsub.channels.length;
  }
})

// psub
Object.defineProperty(Client.prototype, 'psub', {
  get: function getPatternSub() {
    if(!this._pubsub || !this._pubsub.patterns) return 0;
    return this._pubsub.patterns.length;
  }
})

// transaction
Object.defineProperty(Client.prototype, 'transaction', {
  get: function getTransaction() {
    return this._transaction;
  },
  set: function setTransaction(transaction) {
    this._transaction = transaction;
  }
})

// watched keys
Object.defineProperty(Client.prototype, 'watched', {
  get: function getWatched() {
    return this._watched;
  },
  set: function setWatched(watched) {
    this._watched = watched;
  }
})

// pubsub state
Object.defineProperty(Client.prototype, 'pubsub', {
  get: function getPubSub() {
    return this._pubsub;
  },
  set: function setPubSub(pubsub) {
    this._pubsub = pubsub;
  }
})

// client type (class)
Object.defineProperty(Client.prototype, 'type', {
  get: function getType() {
    if(this._pubsub) return CLIENT_TYPE.PUBSUB;
    // slave not yet supported
    return CLIENT_TYPE.NORMAL;
  }
})

Client.prototype.getObject = getObject;
Client.prototype.toString = toString;
Client.prototype.isSlave = isSlave;

module.exports = Client;
