var util = require('util')
  , Constants = require('jsr-constants')
  , StoreError = require('jsr-error')
  , ClientSyntax = StoreError.ClientSyntax
  , IntegerRange = StoreError.IntegerRange
  , NoSuchClient = StoreError.NoSuchClient
  , ClientName = StoreError.ClientName
  , AbstractCommand = require('jsr-exec').Command;

/**
 *  Handler for the CLIENT command.
 */
function Client() {
  AbstractCommand.apply(this, arguments);
}

util.inherits(Client, AbstractCommand);

/**
 *  Respond to the GETNAME subcommand.
 */
function getname(req, res) {
  res.send(null, req.conn.client.name);
}

/**
 *  Respond to the SETNAME subcommand.
 */
function setname(req, res) {
  req.conn.client.name = req.args[0];
  res.send(null, Constants.OK);
}

/**
 *  Respond to the LIST subcommand.
 */
function list(req, res) {
  var s = '';
  this.state.connections.forEach(function(conn) {
    s += conn.toString();
  })
  res.send(null, s);
}

/**
 *  Respond to the PAUSE subcommand.
 */
function pause(req, res) {
  res.send(null, Constants.OK);
  this.state.pause(req.args[0]);
}

/**
 *  Respond to the KILL subcommand.
 */
function kill(req, res) {
  // TODO: implement filters
  var conn;
  // single ip:port
  if(req.args.length === 1) {
    conn = req.args[0];

    // shouldn't get here, however there is a possible
    // race condition whereby the target connection quits
    // between validate and execute
    if(!conn) return res.send(null, null);

    // suicide
    if(conn.id === req.conn.id) {
      conn.write(Constants.OK, function onWrite() {
        conn.close();
      })
    // killing another connection
    }else{
      //console.dir('closing another connection...');
      conn.close(function onClose() {
        res.send(null, Constants.OK);
      })
    }
  }
}

/**
 *  Validate the CLIENT subcommands.
 */
function validate(cmd, args, info) {
  AbstractCommand.prototype.validate.apply(this, arguments);
  var sub = info.command.sub
    , cmd = ('' + sub.cmd)
    , args = sub.args
    , timeout
    , conn;

  if(cmd === Constants.SUBCOMMAND.client.pause.name) {
    timeout = parseInt('' + args[0]);
    // redis returns: 'timeout is' ...
    // but we are not creating another error just yet!
    if(isNaN(timeout)) {
      throw IntegerRange;
    }
    args[0] = timeout;
  }else if(cmd === Constants.SUBCOMMAND.client.kill.name) {
    // kill by ip:port
    if(args.length === 1) {
      conn = this.state.getByAddr('' + args[0]);
      if(!conn) {
        throw NoSuchClient;
      }
      //console.dir('found kill connection');
      args[0] = conn;
    }
  }else if(cmd === Constants.SUBCOMMAND.client.setname.name) {
    // redis appears to allow every thing but whitespace
    if(/\s+/.test(args[0])) {
      throw ClientName;
    }
  }
}

/**
 *  Respond to the CLIENT command.
 */
function execute(req, res) {
  this.delegate(req, res);
}

Client.prototype.validate = validate;
Client.prototype.execute = execute;

// subcommands
Client.prototype.getname = getname;
Client.prototype.setname = setname;
Client.prototype.list = list;
Client.prototype.kill = kill;
Client.prototype.pause = pause;

module.exports = new Client(Constants.MAP.client);
