var util = require('util')
  , Constants = require('jsr-constants')
  , StoreError = require('jsr-error')
  , ConfigParameter = StoreError.ConfigParameter
  , ConfigValue = StoreError.ConfigValue
  , ConfigRewrite = StoreError.ConfigRewrite
  , conf = require('jsr-conf')
  , Configuration = conf.Configuration
  , ConfigKey = conf.Types
  , Runtime = conf.Runtime
  , AbstractCommand = require('jsr-exec').Command;

/**
 *  Handler for the CONFIG command.
 */
function Config() {
  AbstractCommand.apply(this, arguments);
}

util.inherits(Config, AbstractCommand);

/**
 *  Respond to the GET subcommand.
 */
function get(req, res) {
  var ptn = '' + req.args[0]
    , keys = this.pattern.filter(ptn, this.conf.getKeys())
    , list = []
    , i
    , k;
  for(i = 0;i < keys.length;i++) {
    k = keys[i];
    list.push(k, '' + this.conf.get(k));
  }
  res.send(null, list);
}

/**
 *  Respond to the SET subcommand.
 */
function set(req, res) {
  var key = '' + req.args[0]
    , value = req.args[1];

  // already validated the config key/value so should be ok
  // to update the in-memory config, some config verification
  // methods perform side-effects, eg: dir
  this.conf.set(key, value);
  res.send(null, Constants.OK);
  if(key === ConfigKey.DIR) {
    this.log.notice('cwd: %s', process.cwd());
  }
}

/**
 *  Respond to the REWRITE subcommand.
 */
function rewrite(req, res) {

  /* istanbul ignore next: tough to mock fs write error */
  function onError(err) {
    this.conf.removeAllListeners('error');
    this.conf.removeAllListeners('write');
    res.send(err);
  }

  function onWrite() {
    this.conf.removeAllListeners('error');
    this.conf.removeAllListeners('write');
    res.send(null, Constants.OK);
  }

  this.conf.once('error', onError.bind(this));
  this.conf.write(this.conf.getFile(), onWrite.bind(this));
}

/**
 *  Respond to the RESETSTAT subcommand.
 */
function resetstat(req, res) {
  this.stats.reset();
  res.send(null, Constants.OK);
}

/**
 *  Validate the CONFIG subcommands.
 */
function validate(cmd, args, info) {
  AbstractCommand.prototype.validate.apply(this, arguments);
  var sub = info.command.sub
    , cmd = '' + sub.cmd
    , args = sub.args
    , key
    , val
    , verified;

  if(cmd === Constants.SUBCOMMAND.config.set.name) {
    key = '' + args[0];
    val = '' + args[1];

    // test config key is allowed to be set at runtime
    if(!~Runtime.indexOf(key)) {
      throw new ConfigParameter(key);
    }

    // validate input value and coerce to native type
    try {
      val = this.conf.validate(key, [val]);

      // additional validation after type coercion and
      // basic checks, some methods modify the value (hz)
      // although most just throw an error
      verified = this.conf.verify(key, val);
      if(verified !== undefined) {
        val = verified;
      }
      args[1] = val;
    }catch(e) {
      throw new ConfigValue(key, val);
    }
  }else if(cmd === Constants.SUBCOMMAND.config.rewrite.name) {
    if(this.conf.isDefault()) {
      throw ConfigRewrite;
    }
  }
}

/**
 *  Respond to the CONFIG command.
 */
function execute(req, res) {
  this.delegate(req, res);
}

Config.prototype.validate = validate;
Config.prototype.execute = execute;

// subcommands
Config.prototype.get = get;
Config.prototype.set = set;
Config.prototype.rewrite = rewrite;
Config.prototype.resetstat = resetstat;

module.exports = new Config(Constants.MAP.config);
