var util = require('util')
  , events = require('events')
  , net = require('net')
  , Socket = net.Socket
  , Request = require('./request')
  , Response = require('./response')
  , resp = require('jsr-resp')
  , Interpreter = resp.Interpreter
  , Decoder = resp.Decoder
  , Encoder = resp.Encoder
  , itr = new Interpreter()
  , log = require('jsr-log')()
  , Constants = require('jsr-constants')
  , Client = require('./client')
  , STR = Constants.DEC.STR
  , ERR = Constants.DEC.ERR
  , INT = Constants.DEC.INT
  , BST = Constants.DEC.BST
  , ARR = Constants.DEC.ARR;

/**
 *  Encapsulates a client socket connection.
 */
function Connection(socket) {
  events.EventEmitter.call(this);

  // underlying socket
  this._socket = socket;

  // flag whether this connection is paused
  this._paused = false;

  // external or in-process client connection
  this._tcp = (this._socket instanceof Socket);

  // state information
  this._client = new Client(socket);

  this.data = this.data.bind(this);

  this._socket.once('close', this._close.bind(this));
  this._socket.on('data', this.data);

  // using a tcp socket connection
  if(this._tcp) {
    this.encoder = new Encoder({return_buffers: true})
    this.encoder.on('push', onPush.bind(this));
    this.encoder.pipe(this._socket);
    this.decoder = new Decoder({return_buffers: true});
    this.decoder.on('reply', this.decoded.bind(this));
    this.decoder.on('error', this.decoded.bind(this));

    function onSocketError(err) {
      log.warning(err.message);
    }

    this._socket.on('error', onSocketError.bind(this));
  }

  this.onWrite = this.onWrite.bind(this);
}

util.inherits(Connection, events.EventEmitter);

/**
 *  Pause this client connection.
 */
function pause() {
  this._paused = true;
}

/**
 *  Resume this client connection and flush buffered
 *  commands.
 */
function resume() {
  this._paused = false;
  if(this.client._qbuf.length) {
    this.data(this.client._qbuf);
    this.client._qbuf = new Buffer(0);
  }
}

/**
 *  Listener for write events emitted by the database.
 *
 *  Inspects the request object and if the request keys
 *  match any keys being watched by this connection and
 *  the request was from another connection then a conflict
 *  if flagged on a transaction if a transaction exists.
 */
function onWrite(req) {
  if(!this.transaction || !this.watched || req.conn.id === this.id) {
    return false;
  }

  //console.dir(req.keys);
  //console.dir(req.args);

  var keys = req.keys || req.def.getKeys(req.args), i, key;
  for(i = 0;i < keys.length;i++) {
    key = '' + keys[i];
    //console.error('got change with key %s', key);
    //console.error('got change with value %s', value);
    //console.error('got change with conn id %j', req.conn.id);
    if(~this.watched.indexOf(key)) {
      //console.error('setting conflict on %s', keys[i]);
      //console.error('setting conflict from %s', req.conn.id);
      this.transaction.conflict = true;
      return true;
    }
  }
}

/**
 *  Watch an array of keys.
 */
function watch(keys, database) {

  keys = keys.map(function(k) {
    if(k instanceof Buffer) return k.toString();
    return k;
  })

  //console.dir(keys);

  // watch can be called multiple times
  // with the same keys, prevent listener leaks
  this.unwatch(database);

  database.on('write', this.onWrite);
  this.watched = keys;
}

/**
 *  Unwatch all keys.
 */
function unwatch(database) {
  database.removeListener('write', this.onWrite);
  this.watched = null;
}

/**
 *  Read incoming data on the socket connection.
 *
 *  @param buf The incoming buffer or internal request
 *  array.
 */
function data(buf) {
  if(this._tcp) {
    if(this._paused) {
      this.client._qbuf = Buffer.concat(
        [this.client._qbuf, buf], this.client._qbuf.length + buf.length);
      return false;
    }
    this.emit('input', buf.length);
    this._process(buf);
  }else{
    // TODO: handle pausing internal connections!

    // not actually a buffer for internal sockets
    // rather the raw request itself, ie: array with command
    // and arguments
    this.decoded(null, buf);
  }
}

/**
 *  Send this client into a monitor state.
 */
function monitor(server) {
  // cannot receive data anymore we just send it
  this.removeAllListeners('data');

  var listener = this.request.bind(this);

  // keep track of the monitor listeners
  // for clean up
  server._monitor[this.id] = listener;

  // flag on the client for client list flags
  this._client.monitor = true;

  server.on('monitor', listener);
}

/**
 *  Listen for monitor events and write to the client.
 */
function request(str) {
  var res = new Response(this);
  res.send(null, str);
}

/**
 *  Handle an incoming decoded RESP message.
 */
function decoded(err, reply) {
  //console.error('got decoder err %s', err );
  //console.error('got decoder reply %j',reply );
  var res = new Response(this), req, cmd, args;
  if(err) {
    // return decoder errors to the client
    return res.send(err);
  }

  // we should only be handling array request types
  if(!Array.isArray(reply)) {
    return res.send(new Error('bad request, array expected'));
  }else if(!reply.length) {
    return res.send(new Error('bad request, empty array invalid'));
  }

  cmd = reply.shift();

  if(cmd instanceof Buffer) cmd = cmd.toString();

  if(typeof cmd !== 'string') {
    return res.send(new Error('bad request, string command expected'));
  }

  cmd = cmd.toLowerCase();

  // keep track of last command
  this._client.cmd = cmd;

  //console.error('got decoder cmd %s', cmd );
  //console.error('got decoder args %j', reply );

  req = new Request(this, cmd, reply);

  this.emit('command', req, res);
}

/**
 *  Process incoming data and respond accordingly.
 *
 *  @param buf The incoming buffer.
 */
function _process(buf) {
  var b = buf[0], cli, cmd, req, res;

  //console.dir('' + buf);

  // does it look like a RESP message
  if(b === STR
    || b === ERR
    || b === INT
    || b === BST
    || b === ARR
    // already buffered some data, chunked request
    || (this.decoder.isBuffered())) {
    //console.dir('writing buffer to decoder..');
    this.decoder.write(buf);
  }else{
    try {
      cli = itr.interpret(buf.toString());
    }catch(e) {
      // ignore empty command strings
      return;
    }
    cmd = (cli.shift() + '').toLowerCase();
    req = new Request(this, cmd, cli);
    res = new Response(this);
    this.emit('command', req, res);
  }
}

/**
 *  Proxy end to the underlying socket.
 */
function end(cb) {
  if(typeof cb === 'function') {
    this.once('disconnect', cb);
  }

  if(this.tcp) {
    // must call destroy to terminate immediately
    this._socket.destroy();
  }else{
    this._socket.end();
  }
}

/**
 *  Listen for the encoder push event to keep track
 *  of bytes written.
 */
function onPush(buf) {
  //console.dir('push: ' + buf);
  this.emit('output', buf.length);
}

/**
 *  Proxy write to the underlying socket.
 */
function write(data, cb) {
  if(this._tcp) {
    //this.emit('output', data);
    this.encoder.write(data, cb);
  }else{
    this._socket.write.call(this._socket, data, cb);
  }
}

/**
 *  Listener for the socket close event.
 */
function _close(err) {
  this.emit('disconnect', this, this._socket, err);
  this._destroy();
}

/**
 *  Destroy this connection.
 */
function _destroy() {
  this.removeAllListeners();

  this._socket.removeAllListeners();
  this._socket = null;
  this._client = null;
}

/**
 *  Get a string representing this client connection.
 */
function toString() {
  if(this._client) {
    return this._client.toString.apply(this._client, arguments);
  }
  return Object.prototype.toString.call(this);
}

Connection.prototype.toString = toString;

Connection.prototype.watch = watch;
Connection.prototype.onWrite = onWrite;
Connection.prototype.unwatch = unwatch;

Connection.prototype.data = data;
Connection.prototype.monitor = monitor;
Connection.prototype.request = request;
Connection.prototype.end = end;
Connection.prototype.write = write;

// alias end via close
Connection.prototype.close = end;

// TODO: make private
Connection.prototype.decoded = decoded;
Connection.prototype.pause = pause;
Connection.prototype.resume = resume;

Connection.prototype._close = _close;
Connection.prototype._destroy = _destroy;
Connection.prototype._process = _process;

// the underlying socket
Object.defineProperty(Connection.prototype, 'socket', {
  get: function getSocket() {
    return this._socket;
  }
})

// tcp or internal connection
Object.defineProperty(Connection.prototype, 'tcp', {
  get: function getTcp() {
    return this._tcp;
  }
})

// client state information
Object.defineProperty(Connection.prototype, 'client', {
  get: function getClient() {
    return this._client;
  }
})

// connection id
Object.defineProperty(Connection.prototype, 'id', {
  get: function getId() {
    return this._client.id;
  }
})

// db (legacy)
Object.defineProperty(Connection.prototype, 'db', {
  get: function getDatabase() {
    return this._client.db;
  },
  set: function setDatabase(db) {
    this._client.db = db;
  }
})

// authenticated (legacy)
Object.defineProperty(Connection.prototype, 'authenticated', {
  get: function getAuthenticated() {
    return this._client.authenticated;
  },
  set: function getAuthenticated(auth) {
    this._client.authenticated = auth;
  }
})

// transaction (legacy)
Object.defineProperty(Connection.prototype, 'transaction', {
  get: function getTransaction() {
    return this._client.transaction;
  },
  set: function setTransaction(transaction) {
    this._client.transaction = transaction;
  }
})

// watched keys (legacy)
Object.defineProperty(Connection.prototype, 'watched', {
  get: function getWatched() {
    return this._client.watched;
  },
  set: function setWatched(watched) {
    this._client.watched = watched;
  }
})

module.exports = Connection;
