var util = require('util')
  , pkg = require('../package.json')
  , bytes = require('jsr-util').bytes
  , proc = require('getrusage')
  , ConfigKey = require('jsr-conf').Types
  , COMMENT = '#'
  , NA = 'n/a'
  , EOL = '\n'
  , DAY_SECS = (60 * 60 * 24)
  , SERVER = 'server'
  , CLIENTS = 'clients'
  , MEMORY = 'memory'
  , PERSISTENCE = 'persistence'
  , STATS = 'stats'
  , REPLICATION = 'replication'
  , CPU = 'cpu'
  , KEYSPACE = 'keyspace'
  , keys = [
      SERVER,
      CLIENTS,
      MEMORY,
      PERSISTENCE,
      STATS,
      REPLICATION,
      CPU,
      KEYSPACE,
    ]

/**
 *  Encapsulates server information.
 */
function Info(state) {
  this.state = state;
  this.conf = state.conf;
  this.stats = state.stats;
  this.name = pkg.name;
  this.version = pkg.version;
}

/**
 *  Get a header for a section.
 */
function getHeader(name) {
  return util.format('%s %s%s', COMMENT, name, EOL);
}

/**
 *  Get a key value pair.
 */
function getPair(key, value) {
  return util.format('%s:%s%s', key, value, EOL);
}

/**
 *  Get server info section.
 */
function getServer() {
  var o = {}
    , uptime = process.uptime()
  o.version = this.version;
  o.os = process.platform;
  o.arch = process.arch;
  o.process_id = process.pid;

  o.tcp_port = this.state.addr && this.state.addr.port
    ? this.state.addr.port : NA;

  o.uptime_in_seconds = uptime;
  o.uptime_in_days = Math.floor(uptime / DAY_SECS);
  o.hz = this.conf.get(ConfigKey.HZ);
  o.config_file = this.conf.isDefault() ? '' : this.conf.getFile();
  return o;
}

/**
 *  Get clients info section.
 */
function getClients() {
  var o = {};
  o.connected_clients = this.stats.connected;
  return o;
}

/**
 *  Get memory info section.
 */
function getMemory(rusage) {
  var o = {}, mem = process.memoryUsage();
  o.used_memory = mem.heapUsed;
  o.used_memory_human = bytes.humanize(mem.heapUsed, 2, true);
  o.used_memory_rss = mem.rss;
  o.used_memory_peak = rusage.maxrss;
  o.used_memory_peak_human = bytes.humanize(rusage.maxrss, 2, true);
  return o;
}

/**
 *  Get persistence info section.
 */
function getPersistence() {
  var o = {};
  return o;
}

/**
 *  Get stats info section.
 */
function getStats() {
  var o = {};
  o.total_connections_received = this.stats.connections;
  o.total_commands_processed = this.stats.commands;
  o.total_net_input_bytes = this.stats.ibytes;
  o.total_net_output_bytes = this.stats.obytes;

  o.rejected_connections = this.stats.rejected;

  o.expired_keys = this.stats.expired;
  o.keyspace_hits = this.stats.hits;
  o.keyspace_misses = this.stats.misses;

  o.pubsub_channels = this.stats.pubsub_channels;
  o.pubsub_patterns = this.stats.pubsub_patterns;

  return o;
}

/**
 *  Get replication info section.
 */
function getReplication() {
  var o = {};
  return o;
}

/**
 *  Get cpu info section.
 */
function getCpu(rusage) {
  var o = {};
  //console.dir(rusage);
  o.used_cpu_sys = rusage.stime.toFixed(2);
  o.used_cpu_user = rusage.utime.toFixed(2);
  return o;
}

/**
 *  Get keyspace info section.
 */
function getKeyspace() {
  var o = {}
    , i
    , db
    , size;
  for(i in this.state.store.databases) {
    db = this.state.store.databases[i];
    size = db.dbsize();
    if(size) {
      o['db' + i] = util.format('keys=%s,expires=%s', size, db.expiring);
    }
  }
  return o;
}

/**
 *  Get section keys.
 */
function getKeys() {
  return keys;
}

/**
 *  Get an info object.
 */
function getObject(section) {
  var o = {}
    , i
    , k
    , name
    , method
    , sections = section ? [section] : keys
    , rusage = proc.usage();

  for(i = 0;i < sections.length;i++) {
    k = sections[i];
    name = k.charAt(0).toUpperCase() + k.substr(1);
    method = 'get' + name
    o[k] = {
      header: name,
      data: this[method](rusage)
    }
  }
  return o;
}

/**
 *  Get a string representation of the server info suitable
 *  for sending as a bulk string for the INFO command.
 */
function toString(section) {
  if(section && !~keys.indexOf(section)) return '';
  var s = ''
    , k
    , d
    , o = this.getObject(section);
  for(k in o) {
    if(s) s += EOL;
    s += this.getHeader(o[k].header);
    for(d in o[k].data) {
      s += this.getPair(d, o[k].data[d]);
    }
  }
  return s;
}

Info.prototype.getKeys = getKeys;
Info.prototype.getObject = getObject;
Info.prototype.getServer = getServer;
Info.prototype.getClients = getClients;
Info.prototype.getMemory = getMemory;
Info.prototype.getPersistence = getPersistence;
Info.prototype.getStats = getStats;
Info.prototype.getReplication = getReplication;
Info.prototype.getCpu = getCpu;
Info.prototype.getKeyspace = getKeyspace;

Info.prototype.getHeader = getHeader;
Info.prototype.getPair = getPair;
Info.prototype.toString = toString;

module.exports = Info;
