var util = require('util')
  , fs = require('fs')
  , path = require('path')
  , log = require('jsr-log')()
  , net = require('net')
  , Server = net.Server;

/**
 *  Abstract server class.
 *
 *  @param server The main server instance.
 */
function NetworkServer(server) {
  this.server = server;
  function connection(socket) {
    server.connection(socket, this);
  }
  this.on('connection', connection);
}

util.inherits(NetworkServer, Server);

/**
 *  Represents a TCP server.
 */
function TcpServer() {
  NetworkServer.apply(this, arguments);
}

util.inherits(TcpServer, NetworkServer);

/**
 *  Represents a UNIX domain socket server.
 */
function UnixServer() {
  NetworkServer.apply(this, arguments);
}

util.inherits(UnixServer, NetworkServer);

/**
 *  Listen on a UNIX domain socket.
 */
function listen(file, perm) {
  var args = [file]
    , perm = perm || 0700
    , server = this;

  this.file = file;
  this.name = path.basename(file);

  if(this.name === file) {
    file = path.join(process.cwd(), file);
  }

  function onListen() {
    try {
      fs.chmodSync(file, perm);
    }catch(e){}
    log.notice('accepting connections at %s', file);
  }
  args.push(onListen);

  // jump through some hoops for stale sockets
  server.on('error', function (e) {
    if(e.code == 'EADDRINUSE') {
      var client = new net.Socket();
      // handle error trying to talk to server
      client.on('error', function(e) {
        // no other server listening
        if(e.code == 'ECONNREFUSED') {
          fs.unlinkSync(file);
          TcpServer.prototype.listen.apply(server, args);
        }
      });
      client.connect({path: file}, function() {
        // exit with original EADDRINUSE
        throw e;
      });
    }
  });
  Server.prototype.listen.apply(server, args);
}

/**
 *  Close the UNIX domain socket server.
 */
function close() {
  log.notice('removing the unix socket file')
  try {
    fs.unlinkSync(file);
  }catch(e){} // nothing to do at this point

  Server.prototype.close.call(this);
}

UnixServer.prototype.listen = listen;
UnixServer.prototype.close = close;

module.exports = {
  TcpServer: TcpServer,
  UnixServer: UnixServer,
}
