var util = require('util')
  , events = require('events')
  , Pattern = require('jsr-pattern')
  , Constants = require('jsr-constants')
  , PUBSUB = Constants.PUBSUB;

/**
 *  Encapsulates the channels and patterns for a client.
 */
function ClientPubSub() {
  this._channels = [];
  this._patterns = [];
}

Object.defineProperty(ClientPubSub.prototype, 'channels', {
  get: function getChannels() {
    return this._channels;
  }
})

Object.defineProperty(ClientPubSub.prototype, 'patterns', {
  get: function getPatterns() {
    return this._patterns;
  }
})

/**
 *  Encapsulates the pubsub system.
 *
 *  Maps channel identifiers to arrays of client identifiers subscribed
 *  to the given channel.
 *
 *  Clients are given a reference to the channels they are subscribed to,
 *  this global subsystem is responsible for managing the client pubsub
 *  references as well as the global references.
 *
 *  The global instance maps channels as keys on the objects which are arrays
 *  of client identifiers, when a pubsub instance is assigned to a client the
 *  channels and patterns are arrays of the client subscriptions.
 */
function PubSub(stats) {

  // ref to statistics so they can be updated
  // as subscriptions change
  this._stats = stats;

  // normal channels
  this._channels = {};

  // pattern channels
  this._patterns = {};

  this._count = {
    channels: {
      total: 0,       // total number of channels
      subscribers: 0, // total subscribers across all channels
    },
    patterns: {
      total: 0,       // total number of patterns
      subscribers: 0, // total subscribers across all patterns
    }
  }
}

util.inherits(PubSub, events.EventEmitter);

/**
 *  Get the number of subscribers to channels not including
 *  pattern subscribers.
 *
 *  Only applicable for the global pubsub subsystem.
 *
 *  @param channels Array of channels.
 */
//function getChannelSubscriberCount(channels) {
  //var i
    //, k
    //, c = 0
    //, subs;
  //for(i = 0;i < channels.length;i++) {
    //k = channels[i];
    //subs = this._channels[k];
    //if(subs) {
      //c += subs.length;
    //}
  //}
  //return c;
//}

/**
 *  Count number of patterns.
 */
function getPatternCount() {
  return this._count.patterns.total;
}

/**
 *  Count number of channels.
 */
function getChannelCount() {
  return this._count.channels.total;
}

/**
 *  Count number of subscribers to all channels.
 */
function getChannelSubscriberCount() {
  return this._count.channels.subscribers;
}

/**
 *  Count number of subscribers to all patterns.
 *
 *  Suitable for the PUBSUB NUMPAT command.
 */
function getPatternSubscriberCount() {
  return this._count.patterns.subscribers;
}

/**
 *  Get a multi bulk array reply of channels with subscriber counts.
 *
 *  Suitable for the PUBSUB NUMSUB [channel-1 channel-N] command.
 *
 *  @param channels The array of channels.
 */
function getChannelList(channels) {
  var i
    , channel
    , list = [];
  for(i = 0;i < channels.length;i++) {
    channel = channels[i];
    list.push(channel,
      this._channels[channel] ? this._channels[channel].length : 0);
  }
  return list;
}

/**
 *  Get channels matching pattern not including pattern subscribers.
 *
 *  Suitable for the PUBSUB CHANNELS [pattern] command.
 */
function getChannels(pattern) {
  if(!pattern) return Object.keys(this._channels);
  var re = (pattern instanceof RegExp) ? pattern : Pattern.compile(pattern)
    , k
    , list = [];
  for(k in this._channels) {
    if(re.test(k)) {
      list.push(k);
    }
  }
  return list;
}

/**
 *  @private
 *
 *  Subscribe to channel(s).
 *
 *  @param id The command id, either subscribe or psubscribe.
 *  @param prop The client pubsub property either: channels or patterns.
 *  @param target Target subscription object (_channels or _patterns).
 *  @param conn The client connection.
 *  @param channels Array of channels.
 */
function doSubscribe(id, prop, target, conn, channels) {
  var i
    , client = conn.client
    , key                     // string channel id
    , channel                 // array of subscribed clients
    , psub = id === (PUBSUB.PSUBSCRIBE);

  channels = channels.map(function(channel) {
    return '' + channel;
  })

  // keep track of the pubsub on the client
  if(!client.pubsub) {
    client.pubsub = new ClientPubSub();
  }

  // can subscribe to multiple channels at once
  // iterate all the channels the client requested
  for(i = 0;i < channels.length;i++) {
    key = channels[i];
    if(!target[key]) {
      this._count[prop].total++;
      this._stats['pubsub_' + prop]++;
    }
    // reference or create the channel
    channel = target[key] = target[key] || [];
    //if(!channel.length) {
      //channel.type = psub ? PUBSUB.PMESSAGE : PUBSUB.MESSAGE;
    //}
    // compile pattern on subscribe
    if(psub && !channel.re) {
      channel.re = Pattern.compile(key);
    }

    // cannot subscribe twice to the same channel
    if(!~channel.indexOf(client.index)) {
      channel.push(client.index);
      this._count[prop].subscribers++;
    }

    client.pubsub[prop].push(key);

    conn.write([id, key, client.pubsub[prop].length]);
  }
}

/**
 *  @private
 *
 *  Unsubscribe from channel(s).
 *
 *  @param id The command id, either unsubscribe or punsubscribe.
 *  @param prop The client pubsub property either: channels or patterns.
 *  @param target Target subscription object (channels or patterns).
 *  @param conn The client connection.
 *  @param channels Array of channels.
 */
function doUnsubscribe(id, prop, target, conn, channels) {
  var i
    , client = conn.client
    , key
    , channel
    , ind;

  channels = channels.map(function(channel) {
    return '' + channel;
  })

  if(!channels.length) {
    channels = client.pubsub[prop];
  }

  for(i = 0;i < channels.length;i++) {
    key = channels[i];
    channel = target[key];

    // global references
    if(channel) {
      ind = channel.indexOf(client.index);
      if(~ind) {
        // remove from the list of subscribers
        channel.splice(ind, 1);
        // no more subscribers, clean up
        if(!channel.length) {
          delete target[key];
          this._count[prop].total--;
          this._stats['pubsub_' + prop]--;
        }
      }
    }

    // client-level references, array of patterns
    ind = client.pubsub[prop].indexOf(key);
    if(~ind) {
      client.pubsub[prop].splice(ind, 1);
      this._count[prop].subscribers--;
    }

    conn.write([id, key, client.pubsub[prop].length]);
  }

  cleanup(client);
}

/*
 *  Subscribe to channel(s).
 *
 *  It is expected that the command handler has validated
 *  the command arguments.
 *
 *  @param conn The client connection.
 *  @param channels Array of channels.
 */
function subscribe(conn, channels) {
  return this.doSubscribe(
    PUBSUB.SUBSCRIBE,
    PUBSUB.CHANNELS,
    this._channels, conn, channels);
}

/**
 *  Unsubscribe from channel(s).
 *
 *  @param conn The client connection.
 *  @param channels Array of channels.
 */
function unsubscribe(conn, channels) {
  return this.doUnsubscribe(
    PUBSUB.UNSUBSCRIBE,
    PUBSUB.CHANNELS,
    this._channels, conn, channels);
}

/*
 *  Subscribe to pattern channel(s).
 *
 *  It is expected that the command handler has validated
 *  the command arguments.
 *
 *  @param conn The client connection.
 *  @param channels Array of channels.
 */
function psubscribe(conn, channels) {
  return this.doSubscribe(
    PUBSUB.PSUBSCRIBE,
    PUBSUB.PATTERNS,
    this._patterns, conn, channels);
}

/**
 *  Unsubscribe from pattern channel(s).
 *
 *  @param conn The client connection.
 *  @param channels Array of channels.
 */
function punsubscribe(conn, channels) {
  return this.doUnsubscribe(
    PUBSUB.PUNSUBSCRIBE,
    PUBSUB.PATTERNS,
    this._patterns, conn, channels);
}

/**
 *  Publish to channel(s).
 *
 *  @param conn The client publishing the message.
 *  @param channel The channel the message is published to.
 *  @param message The message to publish.
 *
 *  @return The number of clients that received the message.
 */
function publish(conn, channel, message) {
  var c = 0
    , indices = []
    , len
    , k
    , re;

  // exact channel match
  for(k in this._channels) {
    // found the channel
    if(k === channel) {
      indices = this._channels[k].slice(0);
      this.emit(
        PUBSUB.PUBLISH, conn, channel, indices,
          [PUBSUB.MESSAGE, channel, message]);
      break;
    }
  }

  len = indices.length;

  // reset indices for patterns
  // required to prevent channel subscribers
  // from also receiving pmessage events
  indices = [];

  // test against pattern channels
  for(k in this._patterns) {
    // pattern should be pre-compiled
    re = this._patterns[k].re;
    if(re.test(channel)) {
      indices = indices.concat(this._patterns[k]);
      this.emit(
        PUBSUB.PUBLISH, conn, channel, indices,
          [PUBSUB.PMESSAGE, k, channel, message]);
    }
  }

  len += indices.length;

  return len;
}

/**
 *  Clean pubsub references on the client when it has no
 *  more channels or patterns.
 */
function cleanup(client) {
  if(client.pubsub
    && client.pubsub.channels
    && client.pubsub.patterns
    && !client.pubsub.channels.length
    && !client.pubsub.patterns.length) {
    client.pubsub = null;
  }
}

Object.defineProperty(PubSub.prototype, 'channels', {
  get: function getChannels() {
    return this._channels;
  }
})

Object.defineProperty(PubSub.prototype, 'patterns', {
  get: function getPatterns() {
    return this._patterns;
  }
})

PubSub.prototype.subscribe = subscribe;
PubSub.prototype.unsubscribe = unsubscribe;
PubSub.prototype.psubscribe = psubscribe;
PubSub.prototype.punsubscribe = punsubscribe;
PubSub.prototype.publish = publish;
PubSub.prototype.getChannels = getChannels;
PubSub.prototype.getChannelList = getChannelList;

// counters
PubSub.prototype.getChannelCount = getChannelCount;
PubSub.prototype.getPatternCount = getPatternCount;
PubSub.prototype.getChannelSubscriberCount = getChannelSubscriberCount;
PubSub.prototype.getPatternSubscriberCount = getPatternSubscriberCount;

// private
PubSub.prototype.doSubscribe = doSubscribe;
PubSub.prototype.doUnsubscribe = doUnsubscribe;

module.exports = PubSub;
