var util = require('util')
  , net = require('net')
  , events = require('events')
  , path = require('path')
  //, TcpServer = net.Server
  , StoreError = require('jsr-error')
  , MaxClients = StoreError.MaxClients
  , UnknownCommand = StoreError.UnknownCommand
  , ConfigKey = require('jsr-conf').Types
  , Constants = require('jsr-constants')
  , ScriptManager = require('jsr-script').ScriptManager
  , systime = require('jsr-util').time
  , loader = require('jsr-loader')
  , argv = require('./argv')
  , Socket = require('./socket')
  , signals = require('./signals')
  , daemon = require('./daemon')
  //, unix = require('./unix')
  , network = require('./network')
  , TcpServer = network.TcpServer
  , UnixServer = network.UnixServer
  , ServerState = require('./state')
  , Connection = require('./connection')
  , Persistence = require('jsr-persistence')()
  , Shutdown = require('./command/server/shutdown')
  , Script = require('./command/script/script')
  , NOSAVE = Constants.SHUTDOWN.NOSAVE
  , KILL = Constants.SUBCOMMAND.script.kill.name
  , DISCARD = Constants.COMMANDS.DISCARD
  , EXEC = Constants.COMMANDS.EXEC
  , MONITOR = Constants.MAP.monitor.name
  , PORT = Constants.PORT;

/**
 *  Redis server.
 */
function Server(options) {
  //TcpServer.call(this, options);

  var onError = onFatalError.bind(this);

  /* istanbul ignore next: always in test env */
  if(process.env.NODE_ENV !== Constants.TEST) {
    process
      .on('error', onError)
      .on('uncaughtException', onError);
  }

  this._tcp = null;
  this._unix = null;

  // are we in the daemonized process
  this._daemonized = false;

  // server state
  this._state = new ServerState();

  this._state.pubsub.on('publish', this.publish.bind(this));

  // map of  monitor listener functions keyed by connection id
  // we need to clean listeners on disconnect
  this._monitor = {};

  // list of command exec handlers
  this.execs = null;

  this.connection = this.connection.bind(this);

  // listen for new connections
  this.on('connection', this.connection);

  this.command = this.command.bind(this);

  ScriptManager.on('command', this.command);

  // setup signals listener
  signals.bind(this)();
}

util.inherits(Server, events.EventEmitter);

/**
 *  Handle a connection event.
 *
 *  @param socket The underlying client socket.
 *  @param server The TCP or UNIX server.
 */
function connection(socket, server) {
  var conn = new Connection(socket)
    , tcp = (server instanceof TcpServer);

  var maxclients = this.conf.get(ConfigKey.MAXCLIENTS);
  if(maxclients && (this.state.length + 1 > maxclients)) {

    this.state.emit('rejected', conn.client.addr, maxclients);

    this.log.debug('rejected %s, exceeds maxclients %s',
      conn.client.addr, maxclients);

    function onWrite() {
      socket.end();
    }

    return conn.write(MaxClients, onWrite.bind(this));
  }

  if(tcp) {
    this.log.debug('accepted connection from %s', conn.client.addr);
  }else if(server) {
    this.log.debug('accepted connection to %s', server.name);
  }else if(!server) {
    this.log.debug('accepted connection in %s', process.pid);
  }

  var timeout = this.conf.get(ConfigKey.TIMEOUT);
  if(typeof timeout === 'number' && timeout) {
    this.log.debug('setting idle timeout for client #%s to %ss',
      conn.id, timeout);
    // config is seconds, socket want milliseconds
    socket.setTimeout(timeout * 1000, function onTimeout() {
      // TODO: log client connection timeout
      socket.destroy();
    })
  }

  conn.once('disconnect', this.disconnect.bind(this));
  conn.on('command', this.command);
  this._state.addConnection(conn);
}

/**
 *  Emit the monitor event string.
 */
function monitor(req, res) {
  var socket = req.conn.socket
    , args = [req.cmd].concat(req.args)
    , t = systime()
    , str;

  // quote command and args
  args = args.map(function(arg) {
    return '"' + arg + '"';
  })

  // TODO: figure out meaning of ip/host prefix: [0

  // build up the monitor simple string response
  str = new String(
    util.format(Constants.FORMAT.MONITOR,
      t.s, t.m,
      socket.remoteAddress, socket.remotePort, args.join(' ')));

  // emit the event, listening connections will send replies
  // to clients registered for monitor
  this.emit('monitor', str);
}

/**
 *  Dispatches publish messages to target clients.
 */
function publish(conn, channel, indices, message) {
  var i
    , conn;
  for(i = 0;i < indices.length;i++) {
    conn = this._state.getClient(indices[i]);
    if(conn) {
      conn.write(message);
    }
  }
}

/**
 *  Delegate a subcommand to a method declared on the exec instance.
 */
function delegate(req, res) {
  var cmd = req.info.command
    , method;

  if(!cmd.sub) {
    throw new Error('attempt to delegate with no subcommand info');
  }

  method = req.exec[cmd.sub.cmd];

  // subcommand is declared but the command exec handler
  // has not declared the method and is attempting to delegate
  // we treat this as unknown subcommand which allows partial
  // subcommand exec handlers that may call delegate but only
  // implement a subset of the declared subcommands
  if(typeof method !== 'function') {
    return res.send(StoreError.UnknownSubcommand);
  }

  // rewrite the cmd/args
  req.cmd = cmd.sub.cmd;
  req.args = cmd.sub.args;
  method.call(this, req, res);
}

/**
 *  Process an incoming command.
 *
 *  @param req The connection request.
 *  @param res The connection response.
 */
function command(req, res) {

  this.stats.commands++;

  //console.dir(this);

  // guaranteed to have a string command by connection
  var cmd = req.cmd
    // guaranteed to have an arguments array by connection
    , args = req.args

    // select implementation should guarantee a valid db index
    // on the connection
    , db = this.store.getDatabase(req.conn.db)
    // info object used by the validate function
    , exec = this.execs[cmd]
    , info = {
      db: db, conf: this.conf, conn: req.conn, exec: exec, commands: this.execs}
    // should we queue this command in a transaction
    // of just execute it
    , queue = cmd !== DISCARD && cmd !== EXEC
    // flag if command is acceptable when busy
    , acceptable;

  // raw arguments list, before validation
  // used so that the slowlog does not show modified arguments
  req.raw = args.slice(0);

  //console.error('server processing command: %s', cmd);
  //console.error('server processing command: %j', args);
    //

  if(ScriptManager.busy && !req.script) {
    // compare on command instances
    // in case they have been renamed
    acceptable =
      (exec === Shutdown
       && ('' + args[0] || '').toLowerCase() === NOSAVE)
      || (exec === Script
        && ('' + args[0] || '').toLowerCase() === KILL);
    if(!acceptable) return res.send(StoreError.Busy);
  }

  if(exec === false) {
    return res.send(StoreError.UnknownDisabled);
  }else if(!exec) {
    return res.send(new UnknownCommand(cmd));
  }

  try {
    exec.validate.call(this, cmd, args, info);

    // inject references
    req.db = db;
    req.conf = this.conf;
    req.exec = exec;

    // these properties are available on the req
    delete info.db;
    delete info.conn;
    delete info.conf;

    // inject the info object which is decorated with
    // additional useful information on validation
    req.info = info;

    // alias the command definition
    req.def = req.exec.definition;
  }catch(e) {
    // let the transaction decide how to proceed on error
    if(req.conn.transaction) {
      return req.conn.transaction.add(e, req, res);
    }
    return res.send(e);
  }

  if(req.conn.transaction && queue) {
    return req.conn.transaction.add(null, req, res);
  }

  this.execute(exec, req, res);

}

/**
 *  Execute a command.
 *
 *  @param exec The command handler.
 *  @param req The inbound request.
 *  @param res The outbound response.
 */
function execute(exec, req, res) {

  // start timing slow logs before execution
  // validation time is not included nor i/o
  this.state.slowlog.start(req.cmd, req.raw);

  // ensure the slowlog stop is called
  // before encoding and sending the response
  res.before = this.state.slowlog.stop;

  // execute the command handler in the server scope
  exec.execute.call(this, req, res);

  // redis does not display commands that fail
  // validation so it is ok to trigger monitor here
  // currently redis does not appear to respect the
  // skip_monitor flag (EXEC) so we don't either
  //
  // we also call monitor after command execution so
  // that MULTI/EXEC blocks show the correct order
  //
  // and never display the monitor command itself which
  // is consistent with the redis behaviour, however, note
  // that if the monitor command was renamed it would be shown
  // in the monitor output
  if(this.listeners('monitor').length && req.cmd !== MONITOR) {
    this.monitor(req, res);
  }
}

/**
 *  Listener for the connection disconnect event triggered
 *  by the socket close event.
 *
 *  @param conn The connection.
 *  @param socket The underlying socket.
 *  @param err Boolean indicating if a transmission error occured.
 */
function disconnect(conn, socket, err) {
  //console.dir('disconnect');

  // clean monitor listeners
  if(this._monitor[conn.id]) {
    this.removeListener('monitor', this._monitor[conn.id]);
    delete this._monitor[conn.id];
  }

  this._state.removeConnection(conn);
}

/**
 *  Loads command definitions.
 */
function load(opts, cb) {
  if(typeof opts === 'function') {
    cb = opts;
    opts = null;
  }

  var args = opts && opts.args ? opts.args : null
    , defaults = path.normalize(path.join(__dirname, 'command'));
  opts = opts || argv(args);

  if(opts.commands === undefined) {
    opts.commands = [defaults];
  }else if(Array.isArray(opts.commands) && opts.defaults) {
    opts.commands.unshift(defaults);
  }

  function onConfigError(err) {
    //console.dir(err);
    console.error(err.message);
    if(process.env.NODE_ENV !== Constants.TEST) {
      process.exit(1);
    }
  }

  this.conf.once('error', onConfigError.bind(this));

  function onConfigLoad() {
    // if we loaded ok we are done with fatal config errors
    // config set/rewrite need to catch errors now
    this.conf.removeAllListeners('error');

    this.emit('config', this.conf);

    opts.conf = this.conf;

    // spawning the daemon process
    if(this.conf.get(ConfigKey.DAEMONIZE) && !this._daemonized) {
      var pid = daemon(
        path.join(__dirname, 'daemonize.js'), this.conf, opts.args);
      if(cb) cb(pid);
      return;
    }

    this.log.notice('cwd: %s', process.cwd());

    // pass configured timeout to script manager
    ScriptManager.timeout = this.conf.get(ConfigKey.LUA_TIME_LIMIT);

    // spawn now, makes first script executions faster
    // later this should be configurable - allow disabling all
    // scripting
    ScriptManager.spawn();

    // load commands
    return loader(opts, onCommandLoad);
  }

  function onCommandLoad(err, commands) {
    if(err) return cb(err);
    this.execs = commands;
    this.emit('ready', this.execs);

    // inject command list so the script manager
    // can detect write commands
    ScriptManager.commands = commands;

    // loading from rdb files breaks the tests
    if(process.env.NODE_ENV === Constants.TEST) {
      return cb();
    }
    Persistence.load(this.state.store, this.state.conf, onDatabaseLoad);
  }

  function onDatabaseLoad(err, time, diff, version) {
    // load errors are thrown, will cause
    // the process to exit
    if(err) throw err;
    if(err !== false) {
      this.log.notice('db loaded from disk: %s seconds (v%s)', time, version);
      this.emit('database', time, diff, version);
    }
    cb();
  }

  onConfigLoad = onConfigLoad.bind(this);
  onCommandLoad = onCommandLoad.bind(this);
  onDatabaseLoad = onDatabaseLoad.bind(this);

  // load config
  this.conf.load(opts, onConfigLoad);
}

/**
 *  Listen on a port or socket.
 */
function listen() {
  var args = []
    , port = this.conf.get(ConfigKey.PORT)
    , socket = this.conf.get(ConfigKey.UNIXSOCKET)
    , perm = this.conf.get(ConfigKey.UNIXSOCKETPERM)
    , backlog = this.conf.get(ConfigKey.TCP_BACKLOG)
    , cb = typeof arguments[arguments.length - 1] === 'function'
      ? arguments[arguments.length - 1] : null
    , tcp
    , unix;

  if(socket) {
    unix = this._unix = new UnixServer(this);
  }

  // tcp: server.listen(port, [host], [backlog], [callback])
  // unix socket: server.listen(path, [callback])

  // port zero means don't listen on a tcp port
  if(port !== 0) {
    tcp = this._tcp = new TcpServer(this);

    // port comes first
    args.unshift(port);

    // backlog is penultimate argument
    if(typeof backlog === 'number') {
      args.push(backlog);
    }

    //console.dir(args);

    // cb is last argument
    if(cb) args.push(cb);

    function onListen() {
      // for server info, the actual port or socket in use
      // stored on the state, null if listen was never called
      // or if the bind failed, could also be a unix socket
      var addr = tcp.address();
      var port = addr.port;
      this.state.addr = addr;

      // only write pid if we listen
      if(this._daemonized) {
        daemon.pid.write(this.conf);
      }

      this.emit('listen', port);
      this.log.notice('accepting connections on port %s', port);

      // might be listening on a unix socket too
      if(unix) unix.listen(socket, perm);
    }

    tcp.once('listening', onListen.bind(this));
    tcp.listen.apply(tcp, args);
  // just a unix socket
  }else if(unix) {
    unix.listen(socket, perm);
  }
}

/**
 *  Graceful shutdown stop accepting new connections.
 */
function shutdown(opts, cb) {
  if(typeof opts === 'function') {
    cb = opts;
    opts = null;
  }
  opts = opts || {};
  var code = opts.code || 0
    , i = -1
    , snapshots = this.state.conf.get(ConfigKey.SAVE)
    , save = typeof opts.save === 'boolean' ? opts.save : null;

  if(save === null) {
    save = snapshots && snapshots.length > 0;
  }

  /* istanbul ignore next: always in test environment */
  if(process.env.NODE_ENV === Constants.TEST) {
    save = false;
  }

  // cannot save a potentially inconsistent data set
  if(ScriptManager.busy && ScriptManager.written) {
    save = false;
  }

  if(Persistence.loading && save) {
    save = false;
    this.log.warning('save skipped while db load in progress');
  }

  this.state.scheduler.kill();

  function iterate() {
    var conn = this.state.getClient(++i);
    if(!conn) return onClose();
    this.log.verbose('terminating client connection %s', conn.client.id);
    conn.close(function onClose() {
      --i;
      iterate();
    });
  }

  function onShutdownComplete() {
    if(this._daemonized) {
      daemon.pid.remove(this.conf);
    }
    this.log.notice('ready to exit, ciao');
    /* istanbul ignore next: always in test environment */
    if(cb === process.exit
      && process.env.NODE_ENV === Constants.TEST
      && !this._daemonized) {
      return;
    }
    cb(code);
  }

  function onClose() {
    // not saving to disc, explicit nosave
    // or not snapshotting (no save points)
    if(!save) return onShutdownComplete();
    this.log.notice('saving the final rdb snapshot before exiting');
    Persistence.saveSync(this.state.store, this.state.conf);
    onShutdownComplete();
    //Persistence.save(this.state.store, this.state.conf, onShutdownComplete);
  }

  this.log.verbose('not accepting connections')

  // TODO: prevent/handle not running error when not listening!
  // TODO: can possibly test the output of address() to know if we
  // TODO: are listening or not. all logic should be the same just
  // TODO: need to skip the call to close()

  // the close callback can fire before the disconnect listener
  // for the last client connection, so onClose is called when the
  // last connection is actually closed, allows us to assert on
  // client connection length
  if(this._tcp) {
    this._tcp.close();
  }

  if(this._unix) {
    this._unix.close();
  }

  onShutdownComplete = onShutdownComplete.bind(this);
  onClose = onClose.bind(this);
  iterate = iterate.bind(this);
  iterate();
}

/**
 *  Get the server state.
 */
function getState() {
  return this._state;
}

/**
 *  Creates a socket object to be used by a client
 *  within this process.
 */
function createConnection() {

  // create the internal socket object
  var server = new Socket()
    , client = new Socket();

  // pair the sockets
  client.pair = server;
  server.pair = client;

  // register the server side so it is wrapped
  // in a connection and stored as an active connection
  this.connection(server);

  // return the client side of the socket
  return client;
}

Server.prototype.load = load;
Server.prototype.listen = listen;
Server.prototype.monitor = monitor;
Server.prototype.publish = publish;
Server.prototype.connection = connection;
Server.prototype.command = command;
Server.prototype.execute = execute;
Server.prototype.delegate = delegate;
Server.prototype.disconnect = disconnect;
Server.prototype.shutdown = shutdown;

Server.prototype.createConnection = createConnection;
Server.prototype.getState = getState;

// server state property access
Object.defineProperty(Server.prototype, 'state', {
  get: getState
})

// data store
Object.defineProperty(Server.prototype, 'store', {
  get: function getStore() {
    return this._state._store;
  }
})

// data store
Object.defineProperty(Server.prototype, 'info', {
  get: function getInfo() {
    return this._state._info;
  }
})

// shortcut to the configuration
Object.defineProperty(Server.prototype, 'conf', {
  get: function getConfiguration() {
    return this._state._conf;
  }
})

// shortcut to the pattern matcher
Object.defineProperty(Server.prototype, 'pattern', {
  get: function getPattern() {
    return this._state._pattern;
  }
})

// shortcut to the statistics
Object.defineProperty(Server.prototype, 'stats', {
  get: function getStats() {
    return this._state._stats;
  }
})

Object.defineProperty(Server.prototype, 'log', {
  get: function getLogger() {
    return this._state._logger;
  }
})

/**
 *  Handler for all errors on the process.
 */
function onFatalError(err) {
  //console.dir(err.stack);
  this.log.warning(err.message);
  process.exit(1);
}

module.exports = Server;
