var ConfigKey = require('jsr-conf').Types
  // unique id for each log entry
  , id = 0;

/**
 *  Encapsulates a slowlog record.
 */
function SlowLogRecord(cmd, args) {
  this._now = Date.now();
  this._cmd = [cmd].concat(args);

  // start timer value when command execution started
  this._start = process.hrtime();
}

/**
 *  Determine if this log record took longer than
 *  the specified amount.
 */
function isSlowerThan(amount) {
  // always logging
  var diff = process.hrtime(this._start)
    , micro = Math.round(((diff[0] * 1e9) + diff[1]) / 1e3);
  if(amount === 0) return micro;
  if(micro > amount) {
    return micro;
  }
  return -1;
}

/**
 *  Convert this log record to an array, only called if
 *  the command took longer than the configured slower than
 *  amount.
 *
 *  @param id The id for the record.
 *  @param time The execution time in microseconds.
 */
function toArray(id, time) {
  var o = [
    id,
    Math.round(this._now / 1000),
    time,
    this._cmd];
  return o;
}

SlowLogRecord.prototype.isSlowerThan = isSlowerThan;
SlowLogRecord.prototype.toArray = toArray;

/**
 *  Encapsulates the slow log queries storage.
 *
 *  @param conf The server configuration.
 */
function SlowLog(conf) {
  this._conf = conf;

  // log commands slower than this number of microseconds
  // disabled by default
  this._slowerThan = -1;

  // maximum length of the slow log
  this._maxLen = 128;

  // set up log records array
  this.reset();

  // log record for the current executing command
  this._record = null;

  function onConfigLoad() {
    this._slowerThan = conf.get(ConfigKey.SLOWLOG_LOG_SLOWER_THAN);
    this._maxLen = conf.get(ConfigKey.SLOWLOG_MAX_LEN);
  }

  function onSlowerThan(key, value) {
    this._slowerThan = value;
  }

  function onMaxLen(key, value) {
    // minimum record length is one, if a user wishes
    // to disable slow logs they should set slowlog-log-slower-than to -1
    this._maxLen = Math.max(value, 1);
  }

  onConfigLoad = onConfigLoad.bind(this);
  onSlowerThan = onSlowerThan.bind(this);
  onMaxLen = onMaxLen.bind(this);

  this._conf.once('load', onConfigLoad.bind(this));

  this._conf.on(ConfigKey.SLOWLOG_LOG_SLOWER_THAN, onSlowerThan);
  this._conf.on(ConfigKey.SLOWLOG_MAX_LEN, onMaxLen);

  // bind stop here as it is called as a before
  // function on the response
  this.stop = this.stop.bind(this);
}

/**
 *  Start timing command execution.
 */
function start(cmd, args) {
  if(!this.enabled) return false;
  this._record = new SlowLogRecord(cmd, args);
  return this._record;
}

/**
 *  Stop timing current command execution.
 */
function stop() {
  var time;
  if(!this.enabled) return false;
  time = this._record.isSlowerThan(this._slowerThan);
  if(this._record && time > -1) {
    this._records.unshift(this._record.toArray(id++, time));
    if(this._records.length >= this._maxLen) {
      this._records.pop();
    }
  }
}

/**
 *  Get a list of slowlog records.
 *
 *  Suitable for invoking from SLOWLOG GET.
 */
function get(amount) {
  amount = amount !== undefined && amount >= 0 ? amount : 10;
  amount = Math.min(amount, this._records.length);
  return this._records.slice(0, amount);
}

/**
 *  Get the number of slowlog records.
 *
 *  Suitable for invoking from SLOWLOG LEN.
 */
function len() {
  return this._records.length;
}

/**
 *  Reset the slowlog.
 *
 *  Suitable for invoking from SLOWLOG RESET.
 */
function reset() {
  // list of slow log records that were slower
  // than the configured time, implemented as FIFO queue
  // head to tail
  this._records = [];
}

SlowLog.prototype.start = start;
SlowLog.prototype.stop = stop;

SlowLog.prototype.get = get;
SlowLog.prototype.len = len;
SlowLog.prototype.reset = reset;

Object.defineProperty(SlowLog.prototype, 'enabled', {
  get: function getEnabled() {
    return this._slowerThan >= 0;
  }
})

module.exports = SlowLog;
