var util = require('util')
  , events = require('events');

/**
 *  Abstract socket class used to create a bi-directional link between
 *  a server and a client connecting in the same process.
 *
 *  Simulates the `net.Socket` class so that the code flows in exactly
 *  the same way but bypasses encoding, decoding and TCP transmission
 *  overhead.
 */
function SocketProxy() {
  // the socket we are paired with
  this.pair = null;

  // internal sockets have fd -1
  this._handle = {fd: -1};

  // mimic a remote address / remote port
  this.remoteAddress = '0.0.0.0';
  this.remotePort = '0';

  // give the client some time to listen
  // for the connect event
  function onNextTick() {
    this.emit('connect');
  }
  process.nextTick(onNextTick.bind(this));
}

util.inherits(SocketProxy, events.EventEmitter);

/**
 *  Read data from the paired socket.
 */
function read(data) {
  // decoded data should be string primitives
  if(data instanceof String) {
    data = '' + data;
  }
  this.emit('data', data);
}

/**
 *  Write data to the read of the paired socket.
 *
 *  @param data The pair reply object.
 */
function write(data) {
  this.pair.read(data);
}

/**
 *  Close the socket connection.
 */
function end() {
  // can never be a tcp transmission error :)
  this.pair.emit('close', false);
  this.emit('close', false);
  this.pair.destroy();
  this.destroy();
}

/**
 *  Destroy this socket.
 */
function destroy() {
  this.removeAllListeners();
  this.pair = null;
}

SocketProxy.prototype.write = write;
SocketProxy.prototype.read = read;
SocketProxy.prototype.end = end;
SocketProxy.prototype.destroy = destroy;

module.exports = SocketProxy;
