var util = require('util')
  , events = require('events')
  , conf = require('jsr-conf')
  , Configuration = conf.Configuration
  , ConfigName = conf.Types
  , Pattern = require('jsr-pattern')
  , logger = require('jsr-log')
  , Store = require('jsr-store')
  , SlowLog = require('./slowlog')
  , PubSub = require('./pubsub')
  , Scheduler = require('./scheduler')
  , Persistence = require('jsr-persistence')()
  , Stats = require('./stats')
  , Info = require('./info');

/**
 *  Encapsulates the state of the server.
 */
function ServerState() {

  // server configuration
  this._conf = new Configuration();
  this._conf.once('load', this.onConfigLoad.bind(this));

  // the server bind address information
  // if listen() is never called this is null
  this._addr = null;

  // timestamp for the last save of the rdb file
  this._lastsave = Date.now();

  // server logger
  this._logger = logger();

  // periodic scheduler
  this._scheduler = new Scheduler();

  // collection of connections
  this._conn = [];

  // master pattern matcher
  this._pattern = new Pattern();

  // the store - this is where the data lives
  this._store = new Store(this._conf);

  // ensure we have a database at index 0
  // the SELECT command will lazily create databases
  // on demand
  if(this._store.databases[0] === undefined) {
    this._store.getDatabase(0, {pattern: this._pattern});
  }

  // server statistics
  this._stats = new Stats(this);

  // server information
  this._info = new Info(this);

  // global pubsub state
  this._pubsub = new PubSub(this._stats);

  // slow log storage
  this._slowlog = new SlowLog(this._conf);

  function onSave(time, diff, version, file) {
    this._lastsave = Date.now();
    this.log.notice('db saved to disk: %s seconds (%s)', time, file)
  }
  Persistence.on('save', onSave.bind(this));

  process.on('debug', debug.bind(this));
  process.on('tick', expire.bind(this));
}

util.inherits(ServerState, events.EventEmitter);

/**
 *  Pause normal and pubsub clients.
 */
function pause(timeout) {
  var log = this._logger;
  // keep list of matching clients
  var clients = [];
  this._conn.forEach(function(conn) {
    if(!conn.client.isSlave()) {
      clients.push(conn);
      conn.pause();
    }
  })
  function resume() {
    log.notice('resume %s connections', clients.length);
    clients.forEach(function(conn) {
      conn.resume();
    })
  }
  log.notice('pause %s connections for %sms', clients.length, timeout);
  setTimeout(resume, timeout);
}

/**
 *  Scheduler tick event that handles expiring keys.
 *
 *  Passive expiry.
 */
function expire() {
  var dbs = this._store.databases, i;
  for(i in dbs) {
    dbs[i].delExpiredKeys();
  }
}

/**
 *  Listener for the debug log event.
 */
function debug(rusage, memory) {
  var dbs = this._store.databases
    , i
    , db
    , size;
  for(i in dbs) {
    db = dbs[i];
    size = db.dbsize();
    if(size) {
      this._logger.debug('DB %s: %s keys (%s volatile)', i, size, db.expiring);
    }
  }
  this._logger.debug('%s clients connected, %s heap, %s rss bytes',
    this.getLength(), memory.heapUsed, rusage.maxrss);
}

/**
 *  Listener for the config load event.
 */
function onConfigLoad() {
  var file = this._conf.get(ConfigName.LOGFILE);
  this._logger.level(this._conf.get(ConfigName.LOGLEVEL));
  //console.dir('config loaded.');
  if(file && !this._conf.get(ConfigName.DAEMONIZE)) {
    this._logger.stream(file);
  }
  if(this._conf.isDefault()) {
    this._logger.warning('no config file specified, using the default config');
  }

  // TODO: listen for log change events and toggle this
  this._scheduler.startDebugLog();

  this._scheduler.start(this._conf.get(ConfigName.HZ));
}

/**
 *  Get the data store.
 */
function getStore() {
  return this._store;
}

/**
 *  Get the pubsub system.
 */
function getPubSub() {
  return this._pubsub;
}

/**
 *  Get the server information.
 */
function getInfo() {
  return this._info;
}

/**
 *  Get the server configuration.
 */
function getConfiguration() {
  return this._conf;
}

/**
 *  Get the server statistics.
 */
function getStats() {
  return this._stats;
}

/**
 *  Get the logger.
 */
function getLogger() {
  return this._logger;
}

/**
 *  Get the scheduler.
 */
function getScheduler() {
  return this._scheduler;
}

/**
 *  Get the list of client connections.
 */
function getConnections() {
  return this._conn;
}

/**
 *  Get a connection at the specified index.
 */
function getClient(index) {
  return this._conn[index] || null;
}

/**
 *  Get number of connections.
 */
function getLength() {
  return this._conn.length;
}

/**
 *  Attempt to find a connection by ip:port address.
 */
function getByAddr(addr) {
  var i;
  for(i = 0;i < this._conn.length;i++) {
    if(this._conn[i].client.addr === addr) {
      return this._conn[i];
    }
  }
  return null;
}

/**
 *  @private
 *
 *  Add a client connection.
 *
 *  Called by the server to register a connecting client.
 */
function addConnection(conn) {

  this.emit('connect');

  function onInput(len) {
    this._stats.ibytes += len;
  }

  function onOutput(len) {
    this._stats.obytes += len;
  }

  conn.on('input', onInput.bind(this));
  conn.on('output', onOutput.bind(this));

  this._conn.push(conn);

  // pubsub needs to know the client index
  // to be able to quickly find client subscribers
  conn.client.index = this._conn.length - 1;
}

/**
 *  @private
 *
 *  Remove a client connection.
 *
 *  Called by the server when a client disconnects.
 */
function removeConnection(conn) {

  this.emit('disconnect');

  conn.removeAllListeners('input');
  conn.removeAllListeners('output');

  // clean channel subscriptions
  if(conn.client.pubsub) {
    // must slice channels ref
    this._pubsub.unsubscribe(
      conn, conn.client.pubsub.channels.slice(0));
  }

  // clean pattern subscriptions
  // client.pubsub can be null after
  // removing channel subscriptions
  if(conn.client.pubsub) {
    // must slice patterns ref
    this._pubsub.punsubscribe(
      conn, conn.client.pubsub.patterns.slice(0));
  }

  var ind = this._conn.indexOf(conn);
  // remove the connection reference
  // the server disconnect listener is once
  // so we should always have a valid
  // index into the array
  this._conn.splice(ind, 1);
}

// public
ServerState.prototype.getByAddr = getByAddr;
ServerState.prototype.getConfiguration = getConfiguration;
ServerState.prototype.getConnections = getConnections;
ServerState.prototype.getClient = getClient;
ServerState.prototype.getLength = getLength;

Object.defineProperty(ServerState.prototype, 'store', {
  get: getStore
})

Object.defineProperty(ServerState.prototype, 'pubsub', {
  get: getPubSub
})


Object.defineProperty(ServerState.prototype, 'info', {
  get: getInfo
})

Object.defineProperty(ServerState.prototype, 'connections', {
  get: getConnections
})

Object.defineProperty(ServerState.prototype, 'length', {
  get: getLength
})

Object.defineProperty(ServerState.prototype, 'conf', {
  get: getConfiguration
})

Object.defineProperty(ServerState.prototype, 'stats', {
  get: getStats
})

Object.defineProperty(ServerState.prototype, 'log', {
  get: getLogger
})

Object.defineProperty(ServerState.prototype, 'slowlog', {
  get: function getSlowLog() {
    return this._slowlog;
  }
})

Object.defineProperty(ServerState.prototype, 'scheduler', {
  get: getScheduler
})

Object.defineProperty(ServerState.prototype, 'address', {
  get: function getAddr() {
    return this._addr;
  },
  set: function setAddr(addr) {
    this._addr = addr;
  }
})

Object.defineProperty(ServerState.prototype, 'lastsave', {
  get: function getLastSave() {
    return this._lastsave;
  }
})

// internal
ServerState.prototype.addConnection = addConnection;
ServerState.prototype.removeConnection = removeConnection;
ServerState.prototype.onConfigLoad = onConfigLoad;
ServerState.prototype.pause = pause;

module.exports = ServerState;
