var StoreError = require('jsr-error')
  , ExecAbort = StoreError.ExecAbort
  , TransactionError = StoreError.TransactionError
  , Constants = require('jsr-constants')
  , QUEUED = Constants.QUEUED;

/**
 *  Encapsulates the information needed to perform a transaction.
 *
 *  When a connection is in a multi block (transaction) an instance
 *  of this class is assigned to the connection.
 */
function Transaction() {
  // an error that occured during the queuing - signals EXECABORT
  this.error = null;

  // array of objects representing the queued commands
  this.queue = [];

  // flag used to determine whether a watched key conflict exists
  // when true exec will reply with null to signal the client
  // should re-attempt the transaction
  this.conflict = false;
}

/**
 *  Queue a command.
 *
 *  The command and arguments list should already have been
 *  validated to determine if an error would occur (EXECABORT).
 *
 *  @param err A validation error.
 *  @param req The command request.
 *  @param res The command response.
 */
function add(err, req, res) {
  if(err) {
    // TODO: send args length errors

    // transaction errors are not queued
    // but the transaction is left open for more
    // command to be queued
    if(err instanceof TransactionError) {
      return res.send(err);
    }
    // last error
    this.error = err;
  }

  this.queue.push({err: err, req: req, res: res});
  res.send(null, QUEUED);
}

/**
 *  Execute this transaction.
 *
 *  @param scope The server scope.
 *  @param req The request created for the exec command.
 *  @param res The response created for the exec command.
 */
function exec(scope, req, res) {
  var replies = [], item;

  // NOTE: incrby invalid amount will trigger EXECABORT
  // NOTE: wrong number of args are returned to the client
  // NOTE: but other types of validation errors (argument type coercion)
  // NOTE: trigger EXECABORT

  if(this.error) return this.destroy(req, res, ExecAbort);
  if(this.conflict) return this.destroy(req, res, null, null);

  function onReply(err, reply) {
    replies.push(err || reply);
  }

  onReply = onReply.bind(this);

  // fifo queue, maintain order
  while((item = this.queue.shift()) !== undefined) {
    if(item.err) {
      replies.push(item.err);
    }else{
      item.res.cb = onReply;

      // execute via the server so that these command
      // executions are included in slowlog/monitor
      scope.execute(item.req.exec, item.req, item.res);
    }
  }

  // reply and destroy this transaction
  this.destroy(req, res, null, replies);
}

/**
 *  Clear references, unwatch keys and send
 *  a response to the client.
 */
function destroy(req, res, err, reply) {
  req.conn.transaction = null;
  req.conn.unwatch(req.db);
  this.queue = null;
  this.error = null;
  res.send(err, reply);
}

Transaction.prototype.exec = exec;
Transaction.prototype.add = add;
Transaction.prototype.destroy = destroy;

module.exports = Transaction;
